##############################################################################
# Chapter 10                                                                 #
# Sanchez, J. Time Series for Data Scientists, CUP 2023.                        #
# Case Study 1-part III  
# Features three machine learning regression approaches. 
##############################################################################


# Reading in or Download Libraries needed for code

# install.packages("tidyverse")
# install.packages("lubridate")
# install.packages("ggplot2")
# install.packages("Quandl")
# install.packages("dygraphs")
# install.packages("TSstudio")
# install.packages("forecast")
# install.packages("nlme")
# install.packages("h2o")
# install.packages("plotly")
library(tidyverse)
library(lubridate)
library(ggplot2)
library(Quandl)
library(dygraphs)
library(TSstudio)
library(forecast)
library(nlme)
library(h2o) 
library(plotly)

your_api_key <- "qTZLxCLxmbxxxg83GBU2"     #### put your Quandl API key here
Quandl.api_key(your_api_key)

## In RStudio menu, go to Session -> set working directory
#--> to source file location. 

################   Reading in Datasets  ########################
################################################################

edison <- read_csv("edison.csv")[,-1]
head(edison)

colnames(edison) <- c("date","y")
edison$date <- lubridate::ymd_hms(edison$date)
head(edison)

ur <- Quandl(code="FRED/CAURN",     
             type="ts",
             collapse="monthly",      # 2018-Jul to 2022-Mar
             order="asc",
             start_date="2018-07-01",
             end_date="2023-03-31",  
             meta=TRUE)
gdp <- Quandl(code="FRED/CARQGSP", 
              type="ts",
              collapse="quarterly",   # 2018-Q3 to 2021-Q3
              order="asc",
              start_date="2018-07-01",
              end_date="2023-03-31",
              meta=TRUE)
pop <- Quandl(code="FRED/CAPOP",  
              type="ts",
              collapse="annual",      # only years 2018 2019, 2020, 2021 (repeat)
              order="asc",
              start_date="2018-01-01",
              end_date="2023-12-01",
              meta=TRUE)

# Converting Edison Data to a time series and others to xts
edison_xts <- xts(edison$y, edison$date)
gdp_xts <- xts(gdp,order.by = seq(as.Date("2018-07-01"), by = "3 months", length.out = length(gdp)))
pop_xts <- xts(pop,order.by = seq(as.Date("2018-01-01"), by = "year", length.out = length(pop)) )
ur_xts <-  xts(ur, order.by = seq(as.Date("2018-07-01"), by = "month", length.out = length(ur)))

# Looking at their Respective Dygraphs
dygraph(edison_xts, xlab="Date",ylab="Electricity (MWh)", main="California Edison Electricity from 2018-2023")
dygraph(gdp_xts, xlab="Quarter",ylab="GDP", main=" California GDP 2018-2021 (drop Jan 2020 probably due to covid)")
dygraph(pop_xts, xlab="Year",ylab="People (thousands)", main="California Population 2018-2021 (drop probably due to covid and emigration)")
dygraph(ur_xts, xlab="Month",ylab="unemployment rate (%)", main="California Unemployment Rate 2018-2022")

##################### Adding Features ##################################
########################################################################


# (1) adding feature variables from Edison
###########################################
df <- edison # makes a copy of edison for us to add these new features to

df$hour <- lubridate::hour(df$date)
df$day_of_week <- lubridate::wday(df$date, label = TRUE, week_start=1) 
df$month <- lubridate::month(df$date)
df$year <- lubridate::year(df$date)

df$covid <- ifelse(df$date >= "2020-03-01" & df$date <= "2022-12-01",1,0)
head(df);tail(df)
acf(df$y,lag=1500) # We notice 
#that daily seasonality predominates, followed 
# by weekly seasonality

hour_lag <- 1
day_lag <- 24 
week_lag <- 24 * 7 
month_lag <- 24 * 30 
year_lag <- 24 * 365

df <- df %>% mutate(lag_hour= lag(y, n = hour_lag),
                        lag_two=lag(y,n=2),# remove seasonality? Not needed
                        lag_three=lag(y,n=3),#remove seasonality? Not needed
                        lag_four=lag(y,n=4),
                        lag_five=lag(y,n=5),
                        lag_six=lag(y,n=6),
                        lag_halfday=lag(y,n=12),# remove seasonality? Not needed
                        lag_day = lag(y, n = day_lag),  #THIS for daily seas
                        lag_twoday=lag(y,n = 48),# remove seasonality? Perhaps, 
                             # and maybe n=72, since eg Monday, one day past is
                             # Sunday, not a good predictor, two day is Sat, 
                            
                        lag_week = lag(y, n = week_lag),
                        lag_month = lag(y, n = month_lag),
                        lag_year = lag(y,n=year_lag),
                        date = as.Date(date)) %>% 
filter(!is.na(lag_year))

# (2) add features from FRED
##############################

# adding population feature to df
df$pop <- df$year
for( i in seq(length(df$pop))){ 
  if(df$year[i] == "2019"){
    df$pop[i] <- 39437.61
  }else if(df$year[i] == "2020"){
    df$pop[i] <- 39499.74
  }else if(df$year[i] %in% c("2021","2022","2023")){
    df$pop[i] <- 39237.84
  }
}

# adding unemployment rate feature to df
data.ur <- data.frame("date"=seq(as.Date("2018-07-01"),as.Date("2022-03-01"), by = "months"),"ur"=ur)
data.ur$year <- lubridate::year(data.ur$date)
data.ur$month <- lubridate::month(data.ur$date)
df <- left_join(df,data.ur[,c(2:4)],c("month" = "month", "year" = "year"), na_matches = "na")
df$ur[is.na(df$ur)] <- rep(4.2, sum(is.na(df$ur)))

# adding gdp feature to df
gdp.rep <- c()
for( i in seq(length(gdp)+1)){
  if(i == 14){
    gdp.rep <- c(gdp.rep,rep(gdp[13],(3*6)))
  }else{
    gdp.rep <- c(gdp.rep,rep(gdp[i],3))
  }
}
data.gdp <- data.frame("date"=seq(as.Date("2018-07-01"),as.Date("2023-03-31"), by = "months"),"gdp"=gdp.rep)
data.gdp$year <- lubridate::year(data.gdp$date)
data.gdp$month <- lubridate::month(data.gdp$date)
df <- left_join(df,data.gdp[,c(2:4)],c("month" = "month", "year" = "year"), na_matches = "na")
head(df)

################  Checking Variable Utility    #########################
########################################################################

# some labels for the plots 
month.names <- c("Jan", "Feb", "Mar", "Apr", "May", "Jun",
                 "Jul", "Aug", "Sep", "Oct", "Nov", "Dec")
hour.names <- c("12am","1","2","3","4","5","6","7","8","9","10","11","12pm",
                "1","2","3","4","5","6","7","8","9","10","11")

# hour Boxplot
ggplot(df, aes(x = as.factor(hour), y = y )) +
  geom_boxplot() +
  labs(x = "Hour", y = "Electricity (MWh) ") +
  ggtitle("Aggregate Hourly Electricity-- Increase 12-3am, 3am-10am drops\n then 12pm-4 rises and plateaus till 12am") +
  scale_x_discrete(labels = hour.names)

ggplot(df, aes(x = as.factor(hour), y = y)) +
  geom_boxplot() +
  labs(x = "Hour", y = "Electricity (MWh) ") +
  ggtitle("Closer Look: Aggregate Hourly Electricity-- Increase 12-3am,\n3am-10am drops then 12pm-4 rises and plateaus\n till 12am") +
  scale_x_discrete(labels = hour.names)+
  scale_y_continuous(limits = c(8000,17000))

# day of week Boxplot
ggplot(df, aes(x = as.factor(day_of_week), y = y)) +
  geom_boxplot() +
  labs(x = "day of the week", y = "Electricity (MWh)") +
  ggtitle("Aggregate Electricity Day of Week -- \nWeekend and Monday have slight decrease")
ggplot(df, aes(x = as.factor(day_of_week), y = y)) +
  geom_boxplot() +
  labs(x = "day of the week", y = "Electricity (MWh)") +
  ggtitle("Closer Look: Aggregate Electricity Day of Week -- \nLittle Variation Tues-Fri, drops Sat-Sun Increasing Mon")+
  scale_y_continuous(limits = c(7500,15000))

# month Boxplot
ggplot(df, aes(x = as.factor(month), y = y)) +
  geom_boxplot() +
  labs(x = "Month", y = "Electricity (MWh)") +
  ggtitle("Aggregate Monthly Electricity-- Jan-Apr Constant, then Increases \n from May-Aug, then drops till Dec")+
  scale_x_discrete(labels = month.names) + 
  geom_ribbon(data = df[df$month %in% c(6,7,8),], aes(x = month, ymin = -Inf, ymax = Inf), fill = "red", alpha = 0.2)

# year Boxplot
ggplot(df, aes(x = as.factor(year), y = y)) +
  geom_boxplot() +
  labs(x = "year", y = "Electricity (MWh)") +
  ggtitle("Aggregate daily Electricity-- High 2018, slight decrease till 2020,\n slight increase till 2022, drops in 2023")

# Covid Boxplot
ggplot(df, aes(x = as.factor(covid), y = y, group= covid)) +
  geom_boxplot() +
  labs(x = "Covid Time", y = "Electricity (MWh)") +
  ggtitle(" Electricity between Covid and Non-Covid Years: \n very slight increase for non covid years") +
  scale_x_discrete(labels = c("Covid", "Not Covid"))

# analyzing Lags

# Within A Day
ts_lags(edison_xts, lags=c(hour_lag,2,3,4,5,6,12))

# Day Lags
ts_lags(edison_xts, lags=c(day_lag, 24*2, 24*3, 24*4, 24*5, 24*6, week_lag))

# Week, Month, Year Lag
ts_lags(edison_xts, lags=c(month_lag, year_lag))


##################     Modeling     ########################
############################################################

## Preliminary Adjustments

df <- df[,c(1:10,15,20:22)] # subset the dataset to only the variables we will utilize

# we will make variables factors
df$covid <- as.factor(df$covid) 
df$hour <-  as.factor(df$hour)
df$month <- as.factor(df$month) 

class(df$day_of_week) # currently ordered factor, 
#but this class not recognized by h20, must alter to non order
df$day_of_week <- factor(as.character(df$day_of_week)) # creates a non ordered factor
class(df$day_of_week) # factor is non-ordered now
dim(df)

## Making a training, testing set
train_df <- df[1:(nrow(df) - 8760), ]
test_df <- df[(nrow(df) - 8760 + 1):nrow(df), ]

head(df);head(train_df);head(test_df) # looks good
tail(df);tail(train_df);tail(test_df) # looks good

# ###### benchmark the ML models with this model####
# ##################################################

# Multiple linear regression model
lr_var <- lm(y ~ hour  + day_of_week + month
             + year + covid + lag_hour + lag_two + 
               lag_three + lag_day + pop+ ur + gdp - 1, 
             data = train_df)
summary(lr_var)
plot(lr_var)
resid_xts <- xts(residuals(lr_var), train_df$date)
head(resid_xts)
summary(resid_xts)
hist(resid_xts)

# Looking at their Respective Dygraphs
dygraph(resid_xts, xlab="Date",ylab="Electricity (MWh)")%>% dyRangeSelector


# Simple linear regression model
lr_var <- lm(y ~ hour  + day_of_week + month
             + year + covid + lag_hour + lag_two + 
               lag_three + lag_day +lag_week + 
               lag_month+ pop+ ur + gdp - 1, 
                data = train_df)
summary(lr_var)
plot(lr_var)


# removing pop, covid, year
lr_var <- lm(y ~ hour  + day_of_week + month + lag_hour + 
               lag_two + lag_three + lag_day +  ur + gdp - 1, data = train_df)
summary(lr_var)

# predict using Simple linear regression model
test_df$yhat <- predict(lr_var, newdata = test_df)

## Calculate rmse for LR model on the test partition
rmse_lr = sqrt(sum((test_df$y - test_df$yhat)^2)/8760)

## Calculate the short term MAPE score of the LR model on the test partition
mape_lr_short <- mean(abs(test_df$y[1:(730*3)] - test_df$yhat[1:(730*3)]) / test_df$y[1:(730*3)])

## Calculate the long term MAPE score of the LR model on the test partition
mape_lr_long <- mean(abs(test_df$y - test_df$yhat) / test_df$y)

# ##ACF of original
# y <- edison$y
# acf(edison$y, lag=48)
# 
# y_diff <- diff(y, lag=24)
# acf(y_diff, lag=24)
# 
# y_diff2 <- diff(y_diff, lag=24)
# acf(y_diff2, lag=24*7)

## Analyzing Autocorrelation of the LR Residuals
par(mfrow=c(1,1))
sdresiduals=(lr_var$residuals-mean(lr_var$residuals))/sd(lr_var$residuals)
plot(test_df$yhat, type="l", main="forecast")  
plot(sdresiduals, type="l", main="Standardized Residuals of LR Model")  


# ACF and PACF of LR Model Residuals
par(mfrow=c(1,2))
sdresiduals.ts=ts(sdresiduals)
acf(sdresiduals.ts, lag.max=24*7,main="ACF of LM Residuals") # non-stationary
pacf(sdresiduals.ts, lag.max=24*7,main=" PACF of LM Residuals")
par(mfrow=c(1,1))

## Modeling Residuals using auto.arima

# train_df_xts <- xts(train_df$y, train_df$date)
# train_df_xts <- ts(train_df_xts, frequency = 1)
# 
# model2= auto.arima(train_df_xts,   #autoarima requires a ts object
#                    xreg=cbind(model.matrix(~month, train_df)[,-1],
#                               model.matrix(~day_of_week, train_df)[,-1],
#                               model.matrix(~day_of_week, train_df)[,-1],
#                               model.matrix(~covid, train_df)[,-1],
#                               train_df$hour,
#                               train_df$year,
#                               train_df$lag_hour,
#                               train_df$lag_two,
#                               train_df$lag_three,
#                               train_df$lag_day,
#                               train_df$pop,
#                               train_df$ur,
#                               train_df$gdp), 
#                    seasonal=TRUE, 
#                    stationary = TRUE,
#                    stepwise=FALSE, 
#                    approximation=FALSE,
#                    test = c("kpss"))
# summary(model2)
# 
# # Check for White Noise
# checkresiduals(model2) 
# Box.test(model2$residuals, lag=20, type="Ljung") 

## Returning Residuals back into model using gls

# gls.model <- gls(y~hour + day_of_week + month + year + covid + lag_hour
#                  + lag_two + lag_three + lag_day + pop+ ur + gdp - 1, 
#                  data = train_df,
#                  correlation=corARMA(c(model2$coef[1:5]),p=5))
# summary(gls.model)
# 
# # Comparing Residuals Before and after correction
# par(mfrow=c(1,2))
# plot.ts(sdresiduals.ts, 
#         main="Model residuals",ylab="standardized residuals",
#         ylim=c(-3,4))
# abline(h=0) 
# plot(y=residuals(gls.model,type="normalized"),
#      x=as.vector(time(train_df$y)), type="l",
#      ylab=" residuals",
#      xlab="Time",
#      main="Residuals after \nCorrecting autocorrelation")
# abline(h=0)
# par(mfrow=c(1,1))



##  Prepare for ML Methods
h2o.init(max_mem_size = "10G") 
train_h <- as.h2o(train_df)
test_h <- as.h2o(test_df)

# same model used in LR will be used in RF and GB
x=c("hour","day_of_week", "month", "year", "covid", "lag_hour", "lag_two", "lag_three", "lag_day", "pop", "ur", "gdp")
y="y"


###### Simple RF model using 500 trees and 5 folder CV
######################################################

rf_md = h2o.randomForest(training_frame=train_h,
                         nfolds=5,
                         x=x,
                         y=y,
                         ntrees=500,
                         stopping_rounds=10,
                         stopping_metric="RMSE",
                         score_each_iteration=TRUE,
                         stopping_tolerance=0.0001,
                         seed=1234)

## see the contribution of the mode inputs
h2o.varimp_plot(rf_md)

## See model output
rf_md@model$model_summary
rf_md

## learning process of the model as a function of the number of trees
tree_score =rf_md@model$scoring_history$training_rmse
plot_ly(x=seq_along(tree_score), y=tree_score,
        type="scatter", mode="line") %>%
  layout(title="The trained Model Score History",
         yaxis=list(title="RMSE"),
         xaxis=list(title="Num. of Trees"))

## Model performance in test set. Forecasting performance
test_h$pred_rf = h2o.predict(rf_md, test_h)
## transfer h2o data frame to a data.frame object 
test_1= as.data.frame(test_h)

## calculate rmse of the RF model on the test partition
rmse_rf = sqrt(sum((test_1$y -test_1$pred_rf)^2)/8760)

## Calculate the short term MAPE score of the RF model on the test partition
mape_rf_short = mean(abs(test_1$y[1:(730*3)] -test_1$pred_rf[1:(730*3)])/test_1$y[1:(730*3)])

## Calculate the long term MAPE score of the RF model on the test partition
mape_rf_long = mean(abs(test_1$y -test_1$pred_rf)/test_1$y)


## Visualizing RF model performance in the test set
plot_ly(data=test_1) %>%
  add_lines(x=~date, y=~y, name="Actual") %>%
  add_lines(x=~date, y=~yhat, name="Linear Regression", line=
              list(dash="dot")) %>%
  add_lines(x=~date, y=~pred_rf, name="RF", line=
              list(dash="dash")) %>%
  layout(title="Electricity vs. Prediction (lr,RF)",
         yaxis= list(title="MWh"),
         xaxis=list(title="Date"))

# Gradient boosting ##################
######################################

### Train the GB model with the same input used in RF
gbm_md <- h2o.gbm(
                training_frame = train_h,
                nfold=5,
                x=x,
                y=y,
                max_depth=20,
                distribution="gaussian",
                ntrees=500,
                learn_rate=0.1,
                score_each_iteration=TRUE
)

## See model output
gbm_md@model$model_summary
gbm_md

## How important are the model variables in the training (fitting)
h2o.varimp_plot(gbm_md)

## learning process of the model as a function of the number of trees
tree_score =gbm_md@model$scoring_history$training_rmse
plot_ly(x=seq_along(tree_score), y=tree_score,
        type="scatter", mode="line") %>%
  layout(title="The trained Model Score History",
         yaxis=list(title="RMSE"),
         xaxis=list(title="Num. of Trees"))

# test the model's performance on the testing set
test_h$pred_gbm = h2o.predict(gbm_md, test_h)
test_1 <- as.data.frame(test_h)

## calculate rmse of the GB model on the test partition
rmse_gbm = sqrt(sum((test_1$y -test_1$pred_gbm)^2)/8760)

## calculate the short term MAPE Score of the GB model on the test partition
mape_gbm_short = mean(abs(test_1$y[1:(730*3)] -test_1$pred_gbm[1:(730*3)])/test_1$y[1:(730*3)])

## calculate the long term MAPE Score of the GB model on the test partition
mape_gbm_long = mean(abs(test_1$y -test_1$pred_gbm)/test_1$y)

## Visualizing GB model performance in the test set
plot_ly(data=test_1) %>%
  add_lines(x=~date, y=~y, name="Actual") %>%
  add_lines(x=~date, y=~yhat, name="Linear Regression", line=
              list(dash="dot")) %>%
  add_lines(x=~date, y=~pred_gbm, name="GBM", line=
              list(dash="dash"))%>%
  layout(title="Electricity vs. Prediction (lr,GB)",
         yaxis= list(title="MWh"),
         xaxis=list(title="Date"))


# Comparing all three models

## Visualizing All three model performance in the test set
plot_ly(data=test_1) %>%
  add_lines(x=~date, y=~y, name="Actual") %>%
  add_lines(x=~date, y=~yhat, name="Linear Regression", line=
              list(dash="dot")) %>%
  add_lines(x=~date, y=~pred_gbm, name="GBM", line=
              list(dash="dash"))%>%
  add_lines(x=~date, y=~pred_rf, name="RF", line=
              list(dash="dash")) %>%
  layout(title="Electricity vs. Prediction (lr,RF,GB)",
         yaxis= list(title="MWh"),
         xaxis=list(title="Date"))

# Comparing rmse LR-RF-GB
cbind(
  "Benchmark RMSE"=rmse_lr,
  "RF RMSE"=rmse_rf,
  "GB RMSE"= rmse_gbm)

# Comparing Short Term MAPE LR-RF-GB
cbind(
  "Benchmark short-MAPE"=mape_lr_short,
  "RF short-MAPE"=mape_rf_short,
  "GB short-MAPE"= mape_gbm_short)

# Comparing Long Term MAPE LR-RF-GB
cbind(
  "Benchmark long-MAPE"=mape_lr_long,
  "RF long-MAPE"=mape_rf_long,
  "GB longMAPE"= mape_gbm_long) 
